/*
 * Netris -- A free networked version of T*tris
 * Copyright (C) 1994,1995,1996  Mark H. Weaver <mhw@netris.org>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: game.c,v 1.39 1999/05/16 06:56:27 mhw Exp $
 */

#define NOEXT
#include "netris.h"
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <sys/types.h>
#include <netinet/in.h>

static struct option options[] = {
	{ "ascii",		2, 0, 'a' },
	{ "connect",	1, 0, 'c' },
	{ "port",		1, 0, 'p' },
	{ "speed",		1, 0, 'i' },
	{ "level",		1, 0, 'l' },
	{ "spy",		1, 0,  1  },
	{ "seed",		1, 0, 's' },
	{ "robot",		1, 0, 'r' },
	{ "fair-robot",	0, 0, 'F' },
	{ "dropmode",	2, 0, 'd' },
	{ "instadrop",	2, 0, 'D' },
	{ "color",		2, 0, 'C' },
	{ "slowterm",	2, 0, 'S' },
	{ "keys",		1, 0, 'k' },
	{ "rules",		0, 0, 'R' },
	{ "info",		0, 0, 'H' },
	{ "help",		0, 0, 'h' },
	{ 0,			0, 0,  0 }
};

enum { KT_left, KT_right, KT_rotright, KT_rotleft, KT_drop, KT_down,
		KT_toggleSpy, KT_pause, KT_faster, KT_redraw, KT_numKeys };

static char *keyNames[KT_numKeys+1] = {
	"Left", "Right", "RotRight", "RotLeft", "Drop", "Down",
	"ToggleSpy", "Pause", "Faster", "Redraw", NULL };

static char *gameNames[GT_len] = { "OnePlayer", "ClassicTwo" };

static char keyTable[KT_numKeys+1];


ExtFunc void MapKeys(char *newKeys)
{
	int i, k, ch;
	char used[256];
	int errs = 0;

	/* XXX assumptions about ASCII encoding here */
	for (i = k = 0; newKeys[i] && k < KT_numKeys; i++,k++) {
		if (newKeys[i] == '^' && newKeys[i+1])
			keyTable[k] = toupper(newKeys[++i]) - ('A' - 1);
		else
			keyTable[k] = newKeys[i];
	}
	memset(used, 0, sizeof(used));
	for (k = 0; k < KT_numKeys; k++) {
		ch = (unsigned char) keyTable[k];
		if (used[ch]) {
			if (iscntrl(ch) && ch < ' ')
				sprintf(scratch, "Ctrl-%c", ch + ('A' - 1));
			else if (isprint(ch))
				sprintf(scratch, "\"%c\"", ch);
			else
				sprintf(scratch, "0x%X", ch);
			if (!errs)
				fprintf(stderr, "Duplicate key mappings:\n");
			errs++;
			fprintf(stderr, "  %s mapped to both %s and %s\n",
					scratch, keyNames[used[ch]-1], keyNames[k]);
		}
		used[ch] = k + 1;
	}
	if (errs)
		exit(1);
} //MapKeys

ExtFunc void WriteConf(void)
{
	FILE *file_out;

	file_out = fopen(CONFIG_FILE, "w");
	if (file_out == NULL)
		die("Error writing config file");

	fprintf(file_out, "### NETRIS %s Config file ###\n\n", version_string);

	fclose(file_out);
	fprintf(stderr, "Wrote new game configuration to %s\n", CONFIG_FILE);
} //WriteConf

ExtFunc void HandleOption(char tag, char *value)
{
	switch (tag) {
		case 'a':	//ascii
			if (value && !strcasecmp(value, "0")) Game.ascii = 0;
			else Game.ascii = 1;
			break;
		case 'c':	//connect
			initConn = 1;
			hostStr = value;
			break;
		case 'p':	//port
			portStr = value; break;
		case 'i':	//speed (of level 1)
			Game.initspeed = atof(value) * 1e6;
			break;
		case 'l':	//level
			if ((Players[0].score.level = atof(value)) < 1)
				Players[0].score.level = 1;
			if (Players[0].score.level > 15)
				Players[0].score.level = 15;
			break;
		case 1:		//spy
			{
				int i;
				i = atof(value);
				Players[i / 10].spy = i % 10;
			}
			break;
		case 's':	//seed
			Game.seed = atoi(value);
			break;
		case 'r':	//robot
			robotEnable = 1;
			Players[0].flags |= SCF_usingRobot;
			InitRobot(value);
			break;
		case 'F':	//fair robot
			fairRobot = 1;
			Players[0].flags |= SCF_fairRobot;
			break;
		case 'd':	//drop mode
			if (value && !strcasecmp(value, "0")) Players[0].dropmode &= 254;
			else Players[0].dropmode |= 1;
			break;
		case 'D':	//instadrop
			if (value && !strcasecmp(value, "0")) Players[0].dropmode &= 253;
			else Players[0].dropmode |= 2;
			break;
		case 'C':	//color
			if (value && !strcasecmp(value, "1")) Game.color = 1;
			else Game.color = 0;
			break;
		case 'S':	//slowterm
			if (value && !strcasecmp(value, "1")) Game.standout = 1;
			else Game.standout = 0;
			break;
		case 'k':	//keys
			MapKeys(value); break;
		case 'H':	//info
			DistInfo(); exit(0);
		case 'R':	//rules
			Rules(); exit(0);
		case 'h':	//help
			Usage(); exit(0);
		default:
			Usage(); exit(1);
	}
} //HandleParam

ExtFunc void ReadConf(char *filename)
{
	FILE *file_in;
	char buf[513];
	int i;
	char *ch;
	char tag[81], value[81];

	file_in = fopen(filename, "r");
	if (file_in) {
		while (fgets(buf, 512, file_in) != NULL) {
			if ((ch = strchr(buf, '#')))
				*ch = '\0'; // truncate string from # char
			for (i = strlen(buf)-1; i >= 0; i--)
				if (buf[i] == ' ' || buf[i] == '\n' || buf[i] == '\t' || buf[i] == 13)
					buf[i] = '\0';
				else break;

			sscanf(buf, "%80[^= \t] = %80[^\n]", tag, value);
			for (i = 0; options[i].name; i++){
				if (!strcasecmp(options[i].name, tag)) {
					HandleOption(options[i].val, value);
					break;
				}
			}
		}
		fclose(file_in);
	} //read file
	else {
		fprintf(stderr, "Unable to open config file %s.\n", filename);
	} //defaults

} //ReadConf

ExtFunc int StartNewPiece(int scr, Shape *shape)
{
	if (Players[scr].nextShape) {
		Players[scr].curShape = Players[scr].nextShape;
		Players[scr].nextShape = shape;
	}
	else
		Players[scr].curShape = shape;
	Players[scr].curY = Players[scr].boardVisible + 4;
	Players[scr].curX = Players[scr].boardWidth / 2;
	while (!ShapeVisible(Players[scr].curShape, scr,
			Players[scr].curY, Players[scr].curX))
		Players[scr].curY--;
	if (!ShapeFits(Players[scr].curShape, scr,
			Players[scr].curY, Players[scr].curX))
		return 0;
	PlotShape(Players[scr].curShape, scr,
		Players[scr].curY, Players[scr].curX, 1);
	return 1;
}

ExtFunc void OneGame(int scr)
{
	MyEvent event;
	int linesCleared, changed = 0;
	int spied = 0, dropMode = 0;
	int oldPaused = 0, paused = 0, pausedByMe = 0, pausedByThem = 0;
	long pauseTimeLeft;
	int pieceCount = 0;
	int key;
	char *p, *cmd;
	int linevalues[4] = { 40, 100, 400, 1200 }; //= 50*lines! - 10*(lines==1)
	int i;

	Game.speed = Game.initspeed;
	for (i = 1; i < Players[scr].score.level; i++)
		Game.speed /= SPEEDINC;
	if (Game.speed < SPEEDMINIMUM)
		Game.speed = SPEEDMINIMUM;
	ResetBaseTime();
	InitFields();
	for (i = 1; i <= totalPlayers + 1; i++)
		if (Players[i].spy) DrawField(i);
	if (totalPlayers > 0) {
		spied = 1;
	}
	ShowDisplayInfo();
	SetITimer(Game.speed, Game.speed);
	if (robotEnable) {
		RobotCmd(0, "GameType %s\n", gameNames[game]);
		RobotCmd(0, "BoardSize 0 %d %d\n",
			Players[scr].boardVisible, Players[scr].boardWidth);
		for (i = 1; i <= totalPlayers + 1; i++) {
			RobotCmd(0, "BoardSize %d %d %d\n",
				i, Players[i].boardVisible, Players[i].boardWidth);
			RobotCmd(0, "Opponent %d %s %s\n",
				i, Players[i].name, Players[i].host);
			if (Players[i].flags & SCF_usingRobot)
				RobotCmd(0, "OpponentFlag %d robot\n", i);
			if (Players[i].flags & SCF_fairRobot)
				RobotCmd(0, "OpponentFlag %d fairRobot\n", i);
		}
		RobotCmd(0, "TickLength %.3f\n", Game.speed / 1.0e6);
		RobotCmd(0, "BeginGame\n");
		RobotTimeStamp();
	}
	Players[scr].nextShape = ChooseOption(stdOptions);
	while (StartNewPiece(scr, ChooseOption(stdOptions))) {
		ShowScore(scr, Players[scr].score);
		if (robotEnable && !fairRobot)
			RobotCmd(1, "NewPiece %d\n", ++pieceCount);
		if (spied) {
			short shapeNum;
			netint2 data[1];

			shapeNum = ShapeToNetNum(Players[scr].curShape);
			data[0] = hton2(shapeNum);
			SendPacket(scr, NP_newPiece, sizeof(data), data);
		}
		for (;;) {
			changed = RefreshBoard(scr) || changed;
			for (i = 1; i <= totalPlayers+1; i++) if (Players[i].spy && i != me)
				changed = RefreshBoard(i) || changed;
			if (changed) {
				if (!paused) ShowTime();
				refresh();
				changed = 0;
			}
			CheckNetConn();
			switch (WaitMyEvent(&event, EM_any)) {
				case E_alarm:
					if (!MovePiece(scr, -1, 0))
						goto nextPiece;
					else if (spied)
						SendPacket(scr, NP_down, 0, NULL);
					break;
				case E_key:
					p = strchr(keyTable, tolower(event.u.key));
					key = p - keyTable;
					if (robotEnable) {
						RobotCmd(1, "UserKey %d %s\n",
								(int)(unsigned char)event.u.key,
								p ? keyNames[key] : "?");
						break;
					}
					if (!p)
						break;
				keyEvent:
					if (paused && (key != KT_pause) && (key != KT_redraw))
						break;
					switch(key) {
						case KT_left:
							if (MovePiece(scr, 0, -1) && spied)
								SendPacket(scr, NP_left, 0, NULL);
							break;
						case KT_right:
							if (MovePiece(scr, 0, 1) && spied)
								SendPacket(scr, NP_right, 0, NULL);
							break;
						case KT_rotleft:
							if (RotatePiece(scr, 0) && spied)
								SendPacket(scr, NP_rotleft, 0, NULL);
							break;
						case KT_rotright:
							if (RotatePiece(scr, 1) && spied)
								SendPacket(scr, NP_rotright, 0, NULL);
							break;
						case KT_down:
							SetITimer(Game.speed, Game.speed);
							if (MovePiece(scr, -1, 0)) {
								if (spied) SendPacket(scr, NP_down, 0, NULL);
							}
							else goto nextPiece;
							break;
						case KT_drop:
							SetITimer(Game.speed, Game.speed);
							if (DropPiece(scr)) {
								if (spied) SendPacket(scr, NP_drop, 0, NULL);
								if (Players[scr].dropmode == 2) goto nextPiece;
							}
							else goto nextPiece;
							dropMode = Players[scr].dropmode;
							break;
						case KT_pause:
							pausedByMe = !pausedByMe;
							if (game == GT_classicTwo) {
								netint2 data[1];

								data[0] = hton2(pausedByMe);
								SendPacket(scr, NP_pause, sizeof(data), data);
							}
							paused = pausedByMe || pausedByThem;
							if (robotEnable)
								RobotCmd(1, "Pause %d %d\n", pausedByMe,
									pausedByThem);
							ShowPause(pausedByMe, pausedByThem);
							changed = 1;
							break;
						case KT_faster:
							if (game != GT_onePlayer)
								break;
							if ((Game.speed /= SPEEDINC) < SPEEDMINIMUM)
								Game.speed = SPEEDMINIMUM;
							SetITimer(Game.speed, SetITimer(0, 0));
							Players[scr].score.level++;
							ShowScore(scr, Players[scr].score);
							changed = 1;
							break;
						case KT_redraw:
							ScheduleFullRedraw();
//							if (paused)
//								RefreshScreen();
							break;
					}
					if (dropMode && DropPiece(scr) > 0) {
						SetITimer(Game.speed, Game.speed);
						if (spied)
							SendPacket(scr, NP_drop, 0, NULL);
					}
					break;
				case E_robot:
				{
					int num;

					cmd = event.u.robot.data;
					if ((p = strchr(cmd, ' ')))
						*p++ = 0;
					else
						p = cmd + strlen(cmd);
					for (key = 0; keyNames[key]; ++key)
						if (!strcmp(keyNames[key], cmd) &&
								(fairRobot || (1 == sscanf(p, "%d", &num) &&
									num == pieceCount)))
							goto keyEvent;
					if (!strcmp(cmd, "Message")) {
						Message(p);
						changed = 1;
					}
					break;
				}
				case E_net:
					switch(event.u.net.type) {
						case NP_giveJunk:
						{
							netint2 data[2];
							short column;

							memcpy(data, event.u.net.data, sizeof(data[0]));
							column = Random(0, Players[scr].boardWidth);
							data[1] = hton2(column);
							InsertJunk(scr, ntoh2(data[0]), column);
							if (spied)
								SendPacket(scr, NP_insertJunk, sizeof(data), data);
//							Message("Opponent added %d lines");
							break;
						}
						case NP_newPiece:
						{
							short shapeNum;
							netint2 data[1];

							FreezePiece(event.u.net.uid);
							memcpy(data, event.u.net.data, sizeof(data));
							shapeNum = ntoh2(data[0]);
							StartNewPiece(event.u.net.uid, NetNumToShape(shapeNum));
							break;
						}
						case NP_down:
							MovePiece(event.u.net.uid, -1, 0);
							break;
						case NP_left:
							MovePiece(event.u.net.uid, 0, -1);
							break;
						case NP_right:
							MovePiece(event.u.net.uid, 0, 1);
							break;
						case NP_rotleft:
							RotatePiece(event.u.net.uid, 0);
							break;
						case NP_rotright:
							RotatePiece(event.u.net.uid, 1);
							break;
						case NP_drop:
							DropPiece(event.u.net.uid);
							break;
						case NP_clear:
							ClearFullLines(event.u.net.uid);
							break;
						case NP_insertJunk:
						{
							netint2 data[2];

							memcpy(data, event.u.net.data, sizeof(data));
							InsertJunk(event.u.net.uid, ntoh2(data[0]), ntoh2(data[1]));
							break;
						}
						case NP_pause:
						{
							netint2 data[1];

							memcpy(data, event.u.net.data, sizeof(data));
							pausedByThem = ntoh2(data[0]);
							paused = pausedByMe || pausedByThem;
							if (robotEnable)
								RobotCmd(1, "Pause %d %d\n", pausedByMe,
									pausedByThem);
							ShowPause(pausedByMe, pausedByThem);
							changed = 1;
							break;
						}
						default:
							break;
					}
					break;
				case E_lostRobot:
				case E_lostConn:
					goto gameOver;
				default:
					break;
			}
			if (paused != oldPaused) {
				if (paused) {
					PauseTime();
					pauseTimeLeft = SetITimer(0, 0);
				}
				else {
					SetITimer(Game.speed, pauseTimeLeft);
					ResumeTime();
				}
				oldPaused = paused;
			}
		}
	nextPiece:
		dropMode = 0;
		FreezePiece(scr);
		Players[scr].score.drops++;
		Players[scr].score.score++;
		if ((linesCleared = ClearFullLines(scr)) > 0) {
			if (game == GT_onePlayer)
				if ((Players[scr].score.lines / 10) <
						((Players[scr].score.lines + linesCleared) / 10)) {
					if ((Game.speed /= SPEEDINC) < SPEEDMINIMUM)
						Game.speed = SPEEDMINIMUM;
					SetITimer(Game.speed, SetITimer(0, 0));
					Players[scr].score.level++;
				} //level up
			Players[scr].score.score += linevalues[linesCleared];
			Players[scr].score.lines += linesCleared;
			Players[scr].score.adds += linesCleared - (linesCleared < 4);
			if (spied)
				SendPacket(scr, NP_clear, 0, NULL);
		}
		if (game == GT_classicTwo && linesCleared > 1) {
			short junkLines;
			netint2 data[1];

			junkLines = linesCleared - (linesCleared < 4);
			data[0] = hton2(junkLines);
			SendPacket(scr, NP_giveJunk, sizeof(data), data);
		}
	}
gameOver:
	SetITimer(0, 0);
}

ExtFunc int main(int argc, char **argv)
{
	char ch;

	Game.standout = Game.color = 1;
	Game.initspeed = DEFAULT_INTERVAL;
	MapKeys(DEFAULT_KEYS);
	{
		int i;
		char *userName;

		for (i = 0; i <= MAX_SCREENS; i++)
			Players[i].score.level = Players[i].spy = 1;
		if (!(userName = getenv("LOGNAME")) || !userName[0])
			if (!(userName = getenv("USER")) || !userName[0])
				userName = "Anonymous";
		strncpy(Players[0].name, userName, 16); //sizeof(Player.name)
		Players[0].name[16] = 0;
		Players[0].boardWidth = 10;
		Players[0].boardHeight = MAX_BOARD_HEIGHT;
		Players[0].boardVisible = 20;
		Players[0].spy = 1;
		strcpy(Players[0].host, "localhost");
	}

//	if (getopt(argc, argv, "f:") == 'f')
//		ReadConf(optarg);
//	else
	ReadConf(CONFIG_FILE);
	while ((ch = getopt_long(argc, argv,
			"hHRs:r:Fk:c:odDSCap:i:l:", options, NULL)) != -1)
		HandleOption(ch, optarg);
	if (optind < argc) {
		Usage();
		exit(1);
	}
	if (fairRobot && !robotEnable)
		fatal("You can't use the -F option without the -r option");
//	WriteConf();

	InitUtil();
	InitScreens();
	if (initConn) {
		game = GT_classicTwo;
		InitiateConnection(hostStr, portStr);
		HandShake();
		OneGame(me);
	} //client
	else {
		game = GT_onePlayer;
		totalPlayers = 0;
		me = 1;
		memcpy(&Players[me], &Players[0], sizeof(Player));
		OneGame(me);
	} //singleplay
	return 0;
}

/*
 * vi: ts=4 ai
 * vim: noai si
 */
